//
//  main.swift
//  Copy on Write
//
//  Created by Alex Hoppen on 30.07.20.
//  Copyright © 2020 Alex Hoppen. All rights reserved.
//

import Foundation

struct MyCopyOnWriteType {
  /* private */ class Storage {
    var a: Int
    var b: Int
    
    init(a: Int, b: Int) {
      self.a = a
      self.b = b
    }
    
    init(other: Storage) {
      self.a = other.a
      self.b = other.b
    }
  }
  
  /* private */ var storage: Storage
  
  init(a: Int, b: Int) {
    self.storage = Storage(a: a, b: b)
  }
  
  var a: Int {
    get {
      return storage.a
    }
    set {
      if !isKnownUniquelyReferenced(&self.storage) {
        self.storage = Storage(other: self.storage)
      }
      self.storage.a = newValue
    }
  }
  
  var b: Int {
    get {
      return storage.b
    }
    set {
      if !isKnownUniquelyReferenced(&storage) {
        self.storage = Storage(other: storage)
      }
      storage.b = newValue
    }
  }
}

func printMemoryAddress<T>(ofReferenceType value: T) {
  print(unsafeBitCast(value, to: UInt64.self))
}

dynamic func noop(_ x: MyCopyOnWriteType) -> MyCopyOnWriteType { return x }

func performanceTest() {
  var value = MyCopyOnWriteType(a: 0, b: 0)
  let repetitions = 100000000
  let start = Date()
  for _ in 0..<repetitions {
    value = noop(value)
  }
  let end = Date()
  print("\(end.timeIntervalSince(start))s total")
  print("\(end.timeIntervalSince(start) * 1_000_000_000 / Double(repetitions))ns / function call")
}

var struct1 = MyCopyOnWriteType(a: 2, b: 3)
var struct2 = struct1
withUnsafePointer(to: &struct1, { (pointer) in
  print(pointer)
})
withUnsafePointer(to: &struct2, { (pointer) in
  print(pointer)
})
printMemoryAddress(ofReferenceType: struct1.storage)
printMemoryAddress(ofReferenceType: struct2.storage)
struct1.a = 4
printMemoryAddress(ofReferenceType: struct1.storage)
printMemoryAddress(ofReferenceType: struct2.storage)
print(struct1.a)
print(struct2.a)

performanceTest()

// - Go through definition of COW type
// - Show that it has value semantics
// - Print memory addresses of storage before assignment
// - Print memory addresses of storage after assignment
// - Show performance of COW type
