import Foundation

struct MyStruct {
  let a: Int = 1
  let b: Int = 2
}

class MyClass {
  let a: Int = 1
  let b: Int = 2

  init() {}
}

enum MyDirectEnum {
  case singleCase(a: Int, b: Int)
  
  init() {
    self = .singleCase(a: 1, b: 2)
  }
}

indirect enum MyIndirectEnum {
  case singleCase(a: Int, b: Int)
  
  init() {
    self = .singleCase(a: 1, b: 2)
  }
}

dynamic func noop(_ x: MyStruct) -> MyStruct { return x }
dynamic func noop(_ x: MyClass) -> MyClass { return x }
dynamic func noop(_ x: MyDirectEnum) -> MyDirectEnum { return x }
dynamic func noop(_ x: MyIndirectEnum) -> MyIndirectEnum { return x }

func performanceTest() {
  var value = MyIndirectEnum()
  let repetitions = 100000000
  let start = Date()
  for _ in 0..<repetitions {
    value = noop(value)
  }
  let end = Date()
  print("\(end.timeIntervalSince(start))s total")
  print("\(end.timeIntervalSince(start) * 1_000_000_000 / Double(repetitions))ns / function call")
}

performanceTest()

// - Show cost of passing an indirect or direct enum around
