import Foundation

struct MySmallStruct {
  let a: Int = 1
  let b: Int = 2
}

struct MyLargeStruct {
  let a: Int = 1
  let b: Int = 2
  let c: Int = 3
  let d: Int = 4
}

class MySmallClass {
  let a: Int = 1
  let b: Int = 2

  init() {}
}

class MyLargeClass {
  let a: Int = 255
  let b: Int = 127
  let c: Int = 3
  let d: Int = 4

  init() {}
}

enum MySmallDirectEnum {
  case singleCase(a: Int, b: Int)
  
  init() {
    self = .singleCase(a: 1, b: 2)
  }
}

enum MyLargeDirectEnum {
  case singleCase(a: Int, b: Int, c: Int, d: Int)
  
  init() {
    self = .singleCase(a: 1, b: 2, c: 3, d: 4)
  }
}

indirect enum MySmallIndirectEnum {
  case singleCase(a: Int, b: Int)
  
  init() {
    self = .singleCase(a: 1, b: 2)
  }
}

indirect enum MyLargeIndirectEnum {
  case singleCase(a: Int, b: Int, c: Int, d: Int)
  
  init() {
    self = .singleCase(a: 1, b: 2, c: 3, d: 4)
  }
}

protocol MyProtocol {}

extension MySmallStruct: MyProtocol {}
extension MyLargeStruct: MyProtocol {}
extension MySmallClass: MyProtocol {}
extension MyLargeClass: MyProtocol {}
extension MySmallDirectEnum: MyProtocol {}
extension MyLargeDirectEnum: MyProtocol {}
extension MySmallIndirectEnum: MyProtocol {}
extension MyLargeIndirectEnum: MyProtocol {}

dynamic func noop(_ x: MySmallStruct) -> MySmallStruct { return x }
dynamic func noop(_ x: MyLargeStruct) -> MyLargeStruct { return x }
dynamic func noop(_ x: MySmallClass) -> MySmallClass { return x }
dynamic func noop(_ x: MyLargeClass) -> MyLargeClass { return x }
dynamic func noop(_ x: MySmallDirectEnum) -> MySmallDirectEnum { return x }
dynamic func noop(_ x: MyLargeDirectEnum) -> MyLargeDirectEnum { return x }
dynamic func noop(_ x: MySmallIndirectEnum) -> MySmallIndirectEnum { return x }
dynamic func noop(_ x: MyLargeIndirectEnum) -> MyLargeIndirectEnum { return x }
dynamic func noop(_ x: MyProtocol) -> MyProtocol { return x }


dynamic func noopInout(_ x: inout MySmallStruct) { }
dynamic func noopInout(_ x: inout MyLargeStruct) { }
dynamic func noopInout(_ x: inout MySmallClass) { }
dynamic func noopInout(_ x: inout MyLargeClass) { }
dynamic func noopInout(_ x: inout MySmallDirectEnum) { }
dynamic func noopInout(_ x: inout MyLargeDirectEnum) { }
dynamic func noopInout(_ x: inout MySmallIndirectEnum) { }
dynamic func noopInout(_ x: inout MyLargeIndirectEnum) { }
dynamic func noopInout(_ x: inout MyProtocol) { }

func printMemoryAddress<T>(ofReferenceType value: T) {
  print(unsafeBitCast(value, to: UInt64.self))
}

func printMemoryAddress<T>(ofValueType value: inout T) {
  withUnsafePointer(to: &value, { (pointer) in
    print(pointer)
  })
}

func performanceTest() {
  var value = MySmallClass()
//  printMemoryAddress(ofValueType: &value)
  let repetitions = 100000000
  let start = Date()
  for _ in 0..<repetitions {
    // Slow
    value = noop(value)
    // Fast
    noopInout(&value)
  }
  let end = Date()
  print("\(end.timeIntervalSince(start))s total")
  print("\(end.timeIntervalSince(start) * 1_000_000_000 / Double(repetitions))ns / function call")
}

performanceTest()

// - Show performance difference between noop and noopInout
